"""ARver version string module."""

from arver import APPNAME, VERSION
from arver.checksum.properties import libsndfile_version


def version_string() -> str:
    """Return full ARver version string, including libsndfile version."""
    number = VERSION.lstrip('v')
    return f'{APPNAME}-{number} ({libsndfile_version()})'



"""Disc info module for ARver."""

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

import cdio
import discid
import musicbrainzngs
import pycdio

from arver import APPNAME, VERSION, URL
from arver.disc.database import AccurateRipDisc, AccurateRipFetcher
from arver.disc.fingerprint import freedb_id, musicbrainz_id, accuraterip_ids
from arver.disc.utils import frames_to_msf, LEAD_IN_FRAMES

PREGAP_TRACK_NUM = -1
ENHANCED_CD_DATA_TRACK_GAP = 11400


@dataclass
class _Track:
    """Representation of a CD track."""
    num: int
    lba: int
    frames: int
    type: str

    def msf(self) -> str:
        """Return track length as a string in MM:SS.FF format."""
        return frames_to_msf(self.frames)

    def __str__(self):
        numstr = f'{self.num:5d}'

        if self.num == PREGAP_TRACK_NUM:
            numstr = f'{"PGAP":>5s}'
        elif self.type != 'audio':
            numstr = f'{"DATA":>5s}'

        return f'{numstr}    {self.msf():>8s}    {self.frames:6d}\n'


class _DiscType(Enum):
    """
    Disc types relevant to ARver:

    Audio CD (Red Book): a "normal" Compact Disc. Single session which only
    contains audio tracks.

    Mixed Mode CD (Yellow Book): single session, audio tracks and one data
    track exist together. Data track is first. Commonly used as game CDs in
    the 1990s and 2000s.

    Enhanced CD (Blue Book): multisession, first session only contains audio
    tracks, second session contains one data track. The data track usually
    contains music videos or other bonus materials. "Copy Control" CDs fall
    into this category as well.

    Disc ID: not a physical CD, disc data is obtained by disc ID lookup from
    MusicBrainz. Only Audio CDs can be correctly verified using this method.

    Unsupported CD: any edge case not covered by the types listed above.
    """
    UNSUPPORTED = 0
    AUDIO = 1
    MIXED_MODE = 2
    ENHANCED = 3
    DISC_ID = 4


def _have_disc() -> bool:
    """
    Detect if there is a readable disc in drive.

    Use discid instead of pycdio because it's much simpler. Furthermore,
    on error pycdio prints its own message which can't be silenced. discid
    is a dependency anyway, so it's not a problem.

    discid seems to raise error on CDs with no audio tracks. That's good:
    one edge case less to care about.
    """
    try:
        discid.read()
    except discid.DiscError:
        return False

    return True


def _is_multisession(device: cdio.Device) -> bool:
    """Check if disc has more than one session."""
    return device.get_last_session() > 0


def _is_audio_only(track_list: List[_Track]) -> bool:
    """Check if track list only contains audio tracks."""
    types = {track.type for track in track_list}
    return len(types) == 1 and 'audio' in types


def _get_disc_type(device: cdio.Device, track_list: List[_Track]) -> _DiscType:
    """
    Determine disc type based on the following rules:

    No audio tracks -> unsupported disc (this should never happen)
    Single session and only audio tracks -> Audio CD
    Single session and the first track is a data track -> Mixed Mode CD
    Multisession and the last track is a data track -> Enhanced CD
    Anything else -> unsupported disc (no idea what that could be)
    """
    if 'audio' not in [track.type for track in track_list]:
        return _DiscType.UNSUPPORTED

    multisession = _is_multisession(device)

    if not multisession and _is_audio_only(track_list):
        return _DiscType.AUDIO

    if not multisession and track_list[0].type == 'data':
        return _DiscType.MIXED_MODE

    if multisession and track_list[-1].type == 'data':
        return _DiscType.ENHANCED

    return _DiscType.UNSUPPORTED


def _calculate_track_lengths(offset_list: List[int], sectors: int) -> List[int]:
    """
    Return a list of track lengths expressed in CD frames.
    Implementation assumes arguments are obtained by MusicBrainz DiscID query.
    Names of arguments follow respective dictionary keys in MB response.
    """
    shifted = offset_list[1:] + [sectors]
    return [shf - off for shf, off in zip(shifted, offset_list)]


def _get_pregap_track(track_list: List[_Track]) -> Optional[_Track]:
    """
    If the first track doesn't begin immediately after lead in, return a
    pregap track object. Otherwise return None.

    Note that existence of a pregap track doesn't mean it contains any audio.
    Short pregap tracks containing about half a second of silence are quite
    common, but audio data hidden in track one pregap (HTOA) is very rare.
    """
    pregap = None
    frames = track_list[0].lba - LEAD_IN_FRAMES

    if frames > 0:
        pregap = _Track(PREGAP_TRACK_NUM, LEAD_IN_FRAMES, frames, 'audio')

    return pregap


def _fix_last_audio_track(track_list: List[_Track]) -> None:
    """
    Fix length of the last audio track by subtracting the length of gap
    between last audio track and the data track in an enhanced CD. This
    is needed because pycdio includes this gap in sectors count of the
    last audio track.

    Obviously this only makes sense when there is a data track following
    audio tracks, that is in Enhanced CDs.
    """
    for track in reversed(track_list):
        if track.type != 'audio':
            continue

        track.frames -= ENHANCED_CD_DATA_TRACK_GAP
        break


@dataclass
class DiscInfo:
    """
    Representation of Compact Disc properties required for calculation
    of AccurateRip disc IDs and for performing checksum verification.
    """
    pregap: Optional[_Track]
    track_list: List[_Track]
    lead_out: int
    type: _DiscType
    accuraterip_data: Optional[AccurateRipDisc] = field(default=None, init=False)

    def _format_tracklist(self):
        str_ = ''
        str_ += 'track     length     frames\n'
        str_ += '-----    --------    ------\n'

        if self.pregap is not None:
            str_ += str(self.pregap)

        for track in self.track_list:
            str_ += str(track)

        return str_.strip()

    def __str__(self):
        str_ = ''
        str_ += f'AccurateRip disc ID: {self.accuraterip_id()}\n'
        str_ += f'MusicBrainz disc ID: {self.musicbrainz_id()}\n'
        str_ += f'Disc type: {self.disc_type()}\n'
        str_ += '\n' + self._format_tracklist()
        return str_

    @classmethod
    def from_cd(cls) -> 'Optional[DiscInfo]':
        """
        Read disc properties from a physical CD in the default device.
        This is the only way to obtain all necessary information to verify
        each supported CD type (i.e. Audio, Mixed Mode and Enhanced).
        """
        if not _have_disc():
            return None

        device = cdio.Device(driver_id=pycdio.DRIVER_DEVICE)
        first_track_num = device.get_first_track().track
        num_tracks = device.get_num_tracks()
        lead_out = device.get_track(pycdio.CDROM_LEADOUT_TRACK).get_lba()

        track_list = []

        for num in range(first_track_num, num_tracks + 1):
            track = device.get_track(num)
            lba = track.get_lba()

            if num < 99:
                frames = track.get_last_lsn() - track.get_lsn() + 1
            else:
                # track.get_last_lsn() throws an exception for track 99. This looks
                # like a bug in libcdio. Track 99 must be the last track on the CD,
                # so we can use the lead out LBA that we already know to calculate
                # the last LSN of track 99.
                track_last_lsn = lead_out - LEAD_IN_FRAMES - 1
                frames = track_last_lsn - track.get_lsn() + 1

            fmt = track.get_format()
            track_list.append(_Track(num, lba, frames, fmt))

        pregap = _get_pregap_track(track_list)

        disc_type = _get_disc_type(device, track_list)
        if disc_type == _DiscType.UNSUPPORTED:
            return None

        if disc_type == _DiscType.ENHANCED:
            _fix_last_audio_track(track_list)

        return cls(pregap, track_list, lead_out, disc_type)

    @classmethod
    def from_disc_id(cls, disc_id: str) -> 'Optional[DiscInfo]':
        """
        Get disc properties from MusicBrainz by disc ID query. This does not
        provide information about data tracks or the true lead out offset, so
        only Audio CDs can be correctly verified by this method.

        If the disc corresponding to specified disc ID is actually a Mixed
        Mode or Enhanced CD, AccurateRip query may fail, verification may not
        be successful, or resulting confidence values may be lower than what
        would be obtained using a physical CD.
        """
        musicbrainzngs.set_useragent(APPNAME, VERSION, URL)

        try:
            response = musicbrainzngs.get_releases_by_discid(disc_id)
        except musicbrainzngs.ResponseError:
            return None

        lead_out = int(response['disc']['sectors'])
        offsets = response['disc']['offset-list']
        lengths = _calculate_track_lengths(offsets, lead_out)

        track_list = []
        for num, track_data in enumerate(zip(offsets, lengths), start=1):
            track_list.append(_Track(num, *track_data, 'audio'))

        pregap = _get_pregap_track(track_list)
        return cls(pregap, track_list, lead_out, _DiscType.DISC_ID)

    def _audio_tracks(self) -> List[_Track]:
        """Return a list of audio tracks on the CD."""
        return [track for track in self.track_list if track.type == 'audio']

    def _audio_offsets(self) -> List[int]:
        """Return a list of offsets of audio tracks on the CD."""
        return [track.lba for track in self._audio_tracks()]

    def _all_offsets(self) -> List[int]:
        """Return a list of offsets of all tracks on the CD."""
        return [track.lba for track in self.track_list]

    def disc_type(self) -> str:
        """Return a string describing disc type."""
        types = {
            _DiscType.UNSUPPORTED: 'Unsupported CD type',
            _DiscType.AUDIO: 'Audio CD',
            _DiscType.MIXED_MODE: 'Mixed Mode CD',
            _DiscType.ENHANCED: 'Enhanced CD',
            _DiscType.DISC_ID: 'None (disc ID lookup)'
        }
        return types[self.type]

    def musicbrainz_id(self) -> str:
        """Return MusicBrainz disc ID as string."""
        last_audio_track = self._audio_tracks()[-1]
        sectors = last_audio_track.lba + last_audio_track.frames

        # Calculation of MusicBrainz disc IDs requires track offsets from the
        # first CD session. Audio and Mixed Mode CDs are single-session, so
        # calculation requires all track offsets regardless of track type. In
        # Enhanced CDs the first session only contains audio tracks, so by
        # using just the audio track offsets the second session is omitted.
        if self.type == _DiscType.ENHANCED:
            offsets = self._audio_offsets()
        else:
            offsets = self._all_offsets()

        return musicbrainz_id(offsets, sectors)

    def accuraterip_id(self) -> str:
        """Return AccurateRip disc ID as string."""
        num = len(self._audio_tracks())
        freedb = freedb_id(self._all_offsets(), self.lead_out)
        ar1, ar2 = accuraterip_ids(self._audio_offsets(), self.lead_out)
        return f'{num:03d}-{ar1:8s}-{ar2:8s}-{freedb:8s}'

    def fetch_accuraterip_data(self) -> None:
        """
        Download AccurateRip disc data and store AccurateRip responses in
        DiscInfo instance.
        """
        fetcher = AccurateRipFetcher.from_id(self.accuraterip_id())
        self.accuraterip_data = fetcher.fetch()


def get_disc_info(disc_id: Optional[str]) -> Optional[DiscInfo]:
    """
    Helper function for obtaining disc information.
    If disc_id is None read CD TOC from the CD in drive.
    Otherwise get TOC from specified MusicBrainz disc ID.
    """
    if disc_id is None:
        disc_info = DiscInfo.from_cd()
    else:
        disc_info = DiscInfo.from_disc_id(disc_id)

    if disc_info is None:
        if disc_id is None:
            print('Could not read disc. Is there a CD in the drive?')
        else:
            print(f'Could not look up disc ID "{disc_id}", is it correct?')

    return disc_info



"""
Display AccurateRip disc data stored in a dBAR binary file cached
by certain AccurateRip-aware CD rippers (e.g. EAC or Whipper).
"""

import argparse
import sys

from arver.disc.database import AccurateRipParser
from arver.version import version_string


def _parse_args():
    parser = argparse.ArgumentParser(
        description="""Display AccurateRip disc data cached in a dBAR file.""")

    parser.add_argument('dbar_file', nargs=1, help='cached AccurateRip response file')
    parser.add_argument('-v', '--version', action='version', version=version_string())

    return parser.parse_args()


def main():
    args = _parse_args()

    dbar_parser = AccurateRipParser(*args.dbar_file)
    disc = dbar_parser.parse()
    if disc is None:
        print('Failed to parse AccurateRip data file, exiting.')
        sys.exit(1)

    print(disc.summary())
    print()
    print(disc)


if __name__ == '__main__':
    main()



"""Functions for calculating checksums of audio files."""

from typing import Optional, Tuple

from arver.checksum import accuraterip  # type: ignore


def accuraterip_checksums(path, track_no, total_tracks) -> Tuple[int, int]:
    """
    Calculate AccurateRip v1 and v2 checksums of specified file.
    WAV and FLAC formats are supported.

    The track_no and total_tracks arguments are used to recognize
    if the track is the first or the last one on CD. These two
    tracks are treated specially by AccurateRip checksum algorithm.

    Return a pair of checksums (v1, v2) as unsigned integers, or a
    pair of zeros if an error occurred during checksum calculation.
    """
    # pylint: disable=c-extension-no-member
    ar1, ar2 = accuraterip.compute(path, track_no, total_tracks)

    if ar1 is None or ar2 is None:
        return 0x0, 0x0

    return ar1, ar2


def copy_crc(path: str) -> Optional[int]:
    """
    Calculate copy CRC of ripped audio file.
    Returns an unsigned integer on success or None on error.
    """
    # pylint: disable=c-extension-no-member
    try:
        return accuraterip.crc32(path)
    except (OSError, TypeError):
        return None



"""Functions for calculating disc fingerprints (disc IDs) from CD TOC."""

from typing import List, Tuple
from discid import put

from arver.disc.utils import LEAD_IN_FRAMES


def freedb_id(offsets: List[int], leadout: int) -> str:
    """
    Return FreeDB disc ID as 8-digit hex string.

    The first argument is the list of LBA offsets of all tracks from all CD
    sessions, regardless of their types. Specifically, for Enhanced CDs the
    list MUST include the offset of the data track from the second session.

    Including the track from the second session is crucial. discid module
    ignores the second session when reading a physical CD, and returns a
    FreeDB ID incompatible with one used by AccurateRip database.

    The second argument is the LBA offset of the lead out track.
    """
    disc = put(1, len(offsets), leadout, offsets)
    return disc.freedb_id


def musicbrainz_id(offsets: List[int], sectors: int) -> str:
    """
    Return MusicBrainz disc ID as a string.

    The first argument is the list of LBA offsets of all tracks in the first
    CD session. This means that for Mixed Mode CDs the data track offset MUST
    be included, but in Enhanced CDs it MUST NOT, because it belongs to the
    second session. For Enhanced CDs this will effectively be the list of all
    audio track offsets, just like for regular Audio CDs.

    The second argument is the total length of the first CD session in sectors,
    including lead in. This is equivalent to LBA lead out offset in Audio and
    Mixed Mode CDs, but not in Enhanced CDs.
    """
    disc = put(1, len(offsets), sectors, offsets)
    return disc.id


def accuraterip_ids(audio_offsets: List[int], leadout: int) -> Tuple[str, str]:
    """
    Return a pair of AccurateRip disc IDs as 8-digit hex strings.

    The first argument is the list of LBA offsets of audio tracks only. In case
    of Mixed Mode and Enhanced CDs it MUST NOT contain offsets of data tracks
    from any session.

    The second argument is the LBA offset of the lead out track.

    This function is an exception because it internally works on LSN offsets.
    It accepts LBA offsets as arguments to make interfaces of all functions
    defined here consistent. LBA offsets are converted to LSN internally.
    """
    lsn_offsets = [offset - LEAD_IN_FRAMES for offset in audio_offsets]
    lsn_leadout = leadout - LEAD_IN_FRAMES

    id1 = 0
    id2 = 0

    for track_num, offset in enumerate(lsn_offsets, start=1):
        id1 += offset
        id2 += (offset or 1) * track_num

    id1 += lsn_leadout
    id2 += lsn_leadout * (len(lsn_offsets) + 1)

    id1 &= 0xffffffff
    id2 &= 0xffffffff

    return f'{id1:08x}', f'{id2:08x}'



"""Disc-related definitions and helpers also useful elsewhere in ARver."""

LEAD_IN_FRAMES = 150
FRAMES_PER_SECOND = 75


def frames_to_msf(frames: int) -> str:
    """Convert integer number of CD frames to time as mm:ss.ff string."""
    min_ = frames // FRAMES_PER_SECOND // 60
    sec = frames // FRAMES_PER_SECOND % 60
    frm = frames % FRAMES_PER_SECOND
    return f'{min_}:{sec:02d}.{frm:02d}'



"""Utilities for communicating with AccurateRip database."""

import json
import struct
from dataclasses import dataclass
from typing import ClassVar, List, Optional

import requests

from arver import APPNAME, VERSION, URL

FETCH_TIMEOUT_SECONDS = 5
URL_BASE = 'http://www.accuraterip.com/accuraterip/'
USER_AGENT_STRING = f'{APPNAME}/{VERSION} {URL}'


@dataclass
class Header:
    """
    AccurateRip response header. Consists of the number of tracks, two types
    of AccurateRip disc IDs, and a FreeDB disc ID.
    """
    size: ClassVar[int] = 13
    num_tracks: int
    ar_id1: int
    ar_id2: int
    freedb_id: int

    @classmethod
    def from_bytes(cls, data):
        """
        Create Header object from the initial bytes of provided binary data. The data
        is little endian: number of tracks is an unsigned byte, and the three disc IDs
        are unsigned long integers.
        """
        header_bytes = data[:Header.size]
        unpacked = struct.unpack('<BLLL', header_bytes)
        return cls(*unpacked)

    def __str__(self):
        return f'0{self.num_tracks:02d}-{self.ar_id1:08x}-{self.ar_id2:08x}-{self.freedb_id:08x}'


@dataclass
class Track:
    """
    AccurateRip track data. Consists of two AccurateRip checksums and corresponding
    confidence value. The first checksum is either v1 or v2 track checksum (database
    does not indicate which one). The other one is the checksum of frame 450 used for
    offset detection.
    """
    size: ClassVar[int] = 9
    confidence: int
    checksum: int
    checksum_450: int

    @classmethod
    def from_bytes(cls, data):
        """
        Create Track object from the initial bytes of provided binary data. The data is
        little endian: confidence is an unsigned byte, and the two AccurateRip checksums
        are unsigned long integers.
        """
        track_bytes = data[:Track.size]
        unpacked = struct.unpack('<BLL', track_bytes)
        return cls(*unpacked)

    def __str__(self):
        return f'{self.checksum:08x}   (confidence: {self.confidence})'


@dataclass
class Response:
    """
    AccurateRip response decoded from binary format: consists of a Header object
    which stores the number of tracks and three types of disc IDs, and a list of
    Track objects which store two AccurateRip checksums and their confidence.
    """
    header: Header
    tracks: List[Track]

    def __str__(self):
        str_ = []
        str_.append(f'disc ID: {self.header}')
        for num, track in enumerate(self.tracks, start=1):
            str_.append(f'track {num:2d}:   {track}')
        return '\n'.join(str_)


@dataclass
class AccurateRipDisc:
    """
    A collection of all Response objects received from AccurateRip database
    for requested disc.
    """
    responses: List[Response]

    def __str__(self):
        str_ = ''
        for num, response in enumerate(self.responses, start=1):
            str_ += f'AccurateRip response {num}:\n'
            str_ += str(response)
            str_ += '\n\n'

        return str_.strip()

    def __repr__(self):
        return json.dumps(self.make_dict(), indent=2)

    def __len__(self):
        return len(self.responses)

    def _max_confidence(self) -> int:
        return max(track.confidence for response in self.responses for track in response.tracks)

    def summary(self) -> str:
        """Summarize the number of responses and maximum confidence."""
        num = len(self)
        word = 'response' if num == 1 else 'responses'
        return f'Got {num} AccurateRip {word} (max confidence: {self._max_confidence()})'

    def make_dict(self):
        """
        Convert AccurateRipDisc object to a dictionary for easy lookup of checksums
        during file verification. The conversion is "lossy": AccurateRip checksums
        with zero confidence are omitted.

        Resulting dictionary has the following structure:

        {
            "1": {
                "checksum_1": {
                    "confidence": X,
                    "response": Y
                },
                "checksum_2": {
                    ...
                },
                ...
            },
            "2": {
                ...
            },
            ...
        }

        Conversion assumes that checksums other than zero are unique in scope of
        one track. If this is not the case, lower confidence value from later
        response may overwrite higher confidence value from earlier response.
        """
        data = {}

        num_responses = len(self.responses)
        num_tracks = self.responses[0].header.num_tracks

        for trk in range(num_tracks):
            index = trk + 1
            data[index] = {}

            for rsp in range(num_responses):
                track = self.responses[rsp].tracks[trk]

                if track.confidence != 0:
                    data[index][track.checksum] = {
                        'confidence': track.confidence,
                        'response': rsp + 1,
                    }

        return data


# pylint: disable=too-few-public-methods
class AccurateRipFetcher:
    """
    Class for fetching AccurateRip data of a Compact Disc, parsing the
    binary data and representing AccurateRip responses in a usable form.
    """

    def __init__(self, num_tracks: int, ar_id1: str, ar_id2: str, freedb_id: str):
        self._num_tracks = num_tracks
        self._ar_id1 = ar_id1
        self._ar_id2 = ar_id2
        self._freedb_id = freedb_id
        self._raw_bytes = bytes()

    @classmethod
    def from_id(cls, accuraterip_id: str) -> 'AccurateRipFetcher':
        """Create Fetcher object from AccurateRip ID string."""
        parts = accuraterip_id.split('-')
        return cls(int(parts[0]), parts[1], parts[2], parts[3])

    def _make_url(self):
        """Build URL to fetch AccurateRip disc data from."""
        dir_ = f'{self._ar_id1[-1]}/{self._ar_id1[-2]}/{self._ar_id1[-3]}/'
        file_ = f'dBAR-0{self._num_tracks:02d}-{self._ar_id1}-{self._ar_id2}-{self._freedb_id}.bin'
        return URL_BASE + dir_ + file_

    def _shift_bytes(self, num_bytes):
        """Left shift raw disc data by num_bytes (discard initial bytes)."""
        self._raw_bytes = self._raw_bytes[num_bytes:]

    def _validate_header(self, header):
        """Check if AccurateRip response header matches requested disc."""
        if header.num_tracks != self._num_tracks or \
            f'{header.ar_id1:08x}' != self._ar_id1 or \
            f'{header.ar_id2:08x}' != self._ar_id2 or \
            f'{header.freedb_id:08x}' != self._freedb_id:
            raise ValueError(f'Unexpected AccurateRip response header: {header}')

    def _parse_raw_bytes(self):
        """
        Parse AccurateRip binary disc data. The data consists of one or more
        Responses. Each Response consists of a Header and one or more Tracks:

        AccurateRipDisc
        |
        |-- Response1
        |   |-- Header
        |   |-- Track1
        |   |-- Track2
        |   ...
        |   `-- TrackN
        ...
        |
        `-- ResponseN
            |-- Header
            |-- Track1
            |-- Track2
            ...
            `-- TrackN

        All Headers in disc data are expected to be the same, and must match
        the number of tracks and the three disc IDs stored in AccurateRipFetcher
        instance. This means the number of tracks must also be the same in each
        Response.

        Raw disc data is stored in AccurateRipFetcher instance as an array of
        bytes, and is parsed in the following way:

        1. Read Header.size bytes from disc data and create a Header object.
        2. Verify that created Header matches disc data in AccurateRipFetcher instance
           (raise ValueError if it doesn't).
        3. Shift disc data left by Header.size bytes (discard parsed header bytes).
        4. Read the number of tracks from Header. For each track:
            - read Track.size bytes from disc data and create a Track object,
            - shift disc data left by Track.size bytes (discard parsed track bytes).
        5. Create a Response object from the Header and the list of Tracks.
        6. If the size of remaining disc data is greater than zero, repeat steps 1-5.
        7. Return AccurateRipDisc object created from the list of Responses.

        Two exceptions can be raised: ValueError when Header data doesn't match
        the disc info in AccurateRipFetcher instance, and a struct.error when the
        binary disc data cannot be unpacked. Both indicate that disc data acquired
        from AccurateRip is corrupted, and in such case all Responses are discarded
        (they cannot be trusted, even if some of them were successfully parsed).
        """
        responses = []

        while len(self._raw_bytes) > 0:
            header = Header.from_bytes(self._raw_bytes)
            self._validate_header(header)
            self._shift_bytes(Header.size)

            tracks = []
            for _ in range(header.num_tracks):
                tracks.append(Track.from_bytes(self._raw_bytes))
                self._shift_bytes(Track.size)

            responses.append(Response(header, tracks))

        return AccurateRipDisc(responses)

    def fetch(self) -> Optional[AccurateRipDisc]:
        """
        Fetch binary disc data from AccurateRip database. Return an AccurateRipDisc
        object, or None on error.
        """
        try:
            response = requests.get(self._make_url(),
                                    headers={'User-Agent': USER_AGENT_STRING},
                                    timeout=FETCH_TIMEOUT_SECONDS)
            response.raise_for_status()
            self._raw_bytes = response.content
            return self._parse_raw_bytes()
        except (requests.HTTPError, struct.error, ValueError) as error:
            print(error)
            return None


class AccurateRipParser(AccurateRipFetcher):
    """Class for parsing AccurateRip data stored in a cached dBAR file."""

    def __init__(self, path: str):
        # Call parent init with dummy values because we don't know the number of
        # tracks or disc fingerprints. Header validator method is a no-op anyway.
        super().__init__(0, '00000000', '00000000', '00000000')
        self._path = path

    def _validate_header(self, header):
        """
        We only have dummy values for the number of tracks and disc fingerprints,
        so override this method so that it doesn't actually verify these values.
        """

    def parse(self) -> Optional[AccurateRipDisc]:
        """
        Read cached AccurateRip disc data from a file. Return an AccurateRipDisc
        object, or None on error.
        """
        try:
            with open(self._path, mode='rb') as dbar:
                self._raw_bytes = dbar.read()
            return self._parse_raw_bytes()
        except (OSError, struct.error, ValueError) as error:
            print(error)
            return None



"""Representation of a set of ripped CDDA files."""

import os

from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, List, Optional

from arver.checksum.checksum import copy_crc, accuraterip_checksums
from arver.checksum.properties import get_nframes
from arver.disc.info import DiscInfo
from arver.disc.utils import frames_to_msf

AUDIO_FRAMES_PER_CD_SECTOR = 588


class AudioFormatError(Exception):
    """Raised when unsupported audio file (or non-audio file) is read."""


def _shorten_path(path: str, max_length: int = 30) -> str:
    """Shorten long path to an abbreviated file name of specified maximum length."""
    name = os.path.basename(path)
    if len(name) <= max_length:
        return name

    adj = 0 if max_length % 2 != 0 else -1
    midpoint = max_length // 2
    return name[:midpoint + adj] + '~' + name[-midpoint:]


class AudioFile:
    """Audio file to be verified against AccurateRip checksum."""

    def __init__(self, path: str) -> None:
        self.path: str = path

        try:
            self._audio_frames = get_nframes(path)
        except (OSError, TypeError) as exc:
            raise AudioFormatError from exc

        self._arv1: Optional[int] = None
        self._arv2: Optional[int] = None
        self._crc32: Optional[int] = None

    def __str__(self) -> str:
        short_name = _shorten_path(self.path)
        is_cdda = 'yes' if self._is_cd_rip() else 'no'
        cdda_frames = self._audio_frames // AUDIO_FRAMES_PER_CD_SECTOR
        length_msf = frames_to_msf(cdda_frames)

        arv1 = f'{self._arv1:08x}' if self._arv1 is not None else 'unknown'
        arv2 = f'{self._arv2:08x}' if self._arv2 is not None else 'unknown'
        crc32 = f'{self._crc32:08x}' if self._crc32 is not None else 'unknown'

        return f'{short_name:<30s}    {is_cdda:>4s}    ' + \
               f'{length_msf:>8s}    {cdda_frames:>6d}    ' + \
               f'{crc32:>8s}    {arv1:>8s}    {arv2:>8s}'

    def _is_cd_rip(self) -> bool:
        """
        If an audio file was ripped from a CD, its number of frames is
        evenly divisible by the number of audio frames per CD sector.
        """
        return self._audio_frames % AUDIO_FRAMES_PER_CD_SECTOR == 0

    def set_copy_crc(self) -> None:
        """Calculate and set copy CRC."""
        self._crc32 = copy_crc(self.path)

    def set_accuraterip_checksums(self, track_no: int, total_tracks: int) -> None:
        """Calculate and set both types of AccurateRip checksums."""
        self._arv1, self._arv2 = accuraterip_checksums(self.path, track_no, total_tracks)


class _Status(Enum):
    """Possible track verification results."""
    SUCCESS = 0
    FAILED = 1
    NODATA = 2


@dataclass
class TrackVerificationResult:
    """Results of AccurateRip verification of a single track."""
    path: str
    checksum: int
    version: str
    confidence: int
    response: int
    status: _Status

    def _status_string(self) -> str:
        results = {_Status.SUCCESS: 'OK', _Status.FAILED: 'FAILED', _Status.NODATA: 'N/A'}
        return results[self.status]

    def as_table_row(self) -> str:
        """Return string formatted as row suitable for verification summary table."""
        short_name = _shorten_path(self.path)
        status = self._status_string()
        confidence = str(self.confidence) if self.confidence != -1 else '--'
        response = str(self.response) if self.response != -1 else '--'

        return f'{short_name:<30s}    {status:>6s}    {self.checksum:08x}    ' \
               f'{self.version:>4s}    {confidence:>4s}    {response:>4s}'


@dataclass
class DiscVerificationResult:
    """Results of AccurateRip verification of a complete disc."""
    all_ok: ClassVar[str] = 'All tracks verified successfully.'

    ok_nodata: ClassVar[str] = 'All tracks with available checksums verified successfully.'

    some_failed: ClassVar[str] = 'Verification of some tracks failed. Your ' \
    'disc may be damaged.'

    all_failed: ClassVar[str] = 'Verification of all tracks failed. Looks ' \
    'like your disc pressing does not exist in AccurateRip database.'

    tracks: List[TrackVerificationResult]

    def as_table(self) -> str:
        """Format verification results as a table."""

        header = f'{"file name":^30s}    {"result":>6s}    ' \
            f'{"checksum":8s}    {"type":4s}    ' \
            f'{"conf":4s}    {"resp":4s}'.rstrip()
        underline = f'{30*"-"}    {6*"-"}    {8*"-"}    {4*"-"}    {4*"-"}    {4*"-"}'
        text = [header, underline] + [track.as_table_row() for track in self.tracks]
        return '\n'.join(text)

    def summary(self) -> str:
        """Return a string with the description of verification result."""
        num_failed = len([track for track in self.tracks if track.status == _Status.FAILED])
        num_nodata = len([track for track in self.tracks if track.status == _Status.NODATA])

        no_checksum = ''
        if num_nodata != 0:
            plural = 'tracks' if num_nodata > 1 else 'track'
            no_checksum += f'{num_nodata} {plural} not present in AccurateRip database.\n'

        if num_failed == 0 and num_nodata == 0:
            return no_checksum + self.all_ok

        if num_failed == 0 and num_nodata != 0:
            return no_checksum + self.ok_nodata

        if num_failed == len(self.tracks) - num_nodata:
            return no_checksum + self.all_failed

        return no_checksum + self.some_failed


class Rip:
    """This class represents a set of ripped audio files to be verified."""

    def __init__(self, paths: List[str]) -> None:
        self._paths: List[str] = paths
        self._discard_htoa()

        self.tracks: List[AudioFile] = []
        for path in self._paths:
            try:
                self.tracks.append(AudioFile(path))
            except AudioFormatError:
                # ignore non-audio or unsupported audio format
                continue

    def _discard_htoa(self) -> None:
        """Discard paths where file names match commonly used HTOA naming patterns."""
        htoa_patterns = ['track00.wav', 'track00.cdda.wav', 'track00.flac', 'track00.cdda.flac']
        self._paths = [path for path in self._paths if os.path.basename(path) not in htoa_patterns]

    def __str__(self) -> str:
        header = f'{"file name":^30s}    ' + \
            f'{"CDDA":^4s}    {"length":^8s}    {"frames":^6s}    ' + \
            f'{"CRC":^8s}    {"ARv1":^8s}    {"ARv2":^8s}'.rstrip()

        underline = f'{30*"-"}    {4*"-"}    {8*"-"}    {6*"-"}    ' + \
                    f'{8*"-"}    {8*"-"}    {8*"-"}'

        str_ = [header, underline]
        for track in self.tracks:
            str_.append(str(track))
        return '\n'.join(str_)

    def __len__(self) -> int:
        return len(self.tracks)

    def calculate_checksums(self) -> None:
        """
        Iterate file list and calculate copy CRCs and AccurateRip checksums.
        This method must be called before __str__() can be used, otherwise
        all printed checksums will be "unknown".
        """
        total_tracks = len(self.tracks)
        for num, track in enumerate(self.tracks, start=1):
            track.set_copy_crc()
            track.set_accuraterip_checksums(num, total_tracks)

    def verify(self, disc_info: DiscInfo) -> DiscVerificationResult:
        """Verify a set of ripped files against a CD with specified TOC."""
        print(f'Verifying {len(self)} tracks:\n')

        if disc_info.accuraterip_data is None:
            raise ValueError('Cannot verify: missing AccurateRip data!')

        checksums = disc_info.accuraterip_data.make_dict()
        results: List[TrackVerificationResult] = []

        total_tracks = len(self.tracks)
        for num, track in enumerate(self.tracks, start=1):
            crc32 = copy_crc(track.path)
            ar1, ar2 = accuraterip_checksums(track.path, num, total_tracks)

            print(f'Track {num}:')
            print(f'\tPath: {track.path}')
            print(f'\tCopy CRC: {crc32:08x}')

            if len(checksums[num]) == 0:
                results.append(
                    TrackVerificationResult(track.path, ar2, 'ARv2', -1, -1, _Status.NODATA))
                print('\tAccurateRip: no checksums available for this track')
                continue

            if ar2 in checksums[num]:
                conf = checksums[num][ar2]['confidence']
                resp = checksums[num][ar2]['response']
                print(f'\tAccurateRip: {ar2:08x} (ARv2), confidence {conf}, response {resp}')
                results.append(
                    TrackVerificationResult(track.path, ar2, 'ARv2', conf, resp, _Status.SUCCESS))
                continue

            if ar1 in checksums[num]:
                conf = checksums[num][ar1]['confidence']
                resp = checksums[num][ar1]['response']
                print(f'\tAccurateRip: {ar1:08x} (ARv1), confidence {conf}, response {resp}')
                results.append(
                    TrackVerificationResult(track.path, ar1, 'ARv1', conf, resp, _Status.SUCCESS))
                continue

            print(f'\tAccurateRip: {ar2:08x} (ARv2) - no match!')
            results.append(TrackVerificationResult(track.path, ar2, 'ARv2', -1, -1, _Status.FAILED))

        return DiscVerificationResult(results)



from setuptools import setup, find_packages, Extension
from arver import APPNAME, VERSION, URL

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

setup(name=APPNAME,
      version=VERSION,
      description='Application for verifying ripped audio files using AccurateRip database.',
      long_description=readme,
      long_description_content_type="text/markdown",
      author='arcctgx',
      author_email='arcctgx@o2.pl',
      url=URL,
      license='GPLv3',
      classifiers=[
          'Development Status :: 5 - Production/Stable', 'Environment :: Console',
          'Intended Audience :: End Users/Desktop',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English', 'Operating System :: POSIX :: Linux',
          'Programming Language :: C', 'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3.7', 'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9', 'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Topic :: Multimedia :: Sound/Audio :: Analysis',
          'Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Ripping'
      ],
      python_requires='>=3.7',
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'arver = arver.arver_main:main', 'arver-discinfo = arver.arver_discinfo:main',
              'arver-ripinfo = arver.arver_ripinfo:main',
              'arver-bin-parser = arver.arver_bin_parser:main'
          ]
      },
      ext_modules=[
          Extension('arver.checksum.accuraterip',
                    sources=['arver/checksum/accuraterip.c'],
                    libraries=['sndfile', 'z'])
      ],
      install_requires=['discid', 'musicbrainzngs', 'pycdio', 'requests'])



"""
Display information about compact disc in drive, or derived from
MusicBrainz disc ID. Download and display AccurateRip data of the CD.
"""

import argparse
import sys

from arver.disc.info import get_disc_info
from arver.version import version_string


def _parse_args():
    parser = argparse.ArgumentParser(
        description="""Display information about compact disc in drive, or
        derived from MusicBrainz disc ID. Fetch and display AccurateRip data
        of the disc.""")

    parser.add_argument('-i',
                        '--disc-id',
                        metavar='disc_id',
                        nargs='?',
                        help='get disc TOC from MusicBrainz by disc ID')

    parser.add_argument('-v', '--version', action='version', version=version_string())

    return parser.parse_args()


def main():
    args = _parse_args()
    disc = get_disc_info(args.disc_id)

    if disc is None:
        print('Failed to get disc info, exiting.')
        sys.exit(1)

    print(disc)
    print()

    disc.fetch_accuraterip_data()
    if disc.accuraterip_data is None:
        print('Failed to download AccurateRip data, exiting.')
        sys.exit(2)

    print(disc.accuraterip_data.summary())
    print()
    print(disc.accuraterip_data)


if __name__ == '__main__':
    main()



"""
Functions for getting properties of supported audio files,
or of libsndfile library itself.
"""

# pylint: disable=c-extension-no-member

from arver.checksum import accuraterip


def get_nframes(path: str) -> int:
    """
    Return the number of frames in a supported audio file, or None
    on error. A frame is a set of samples, one sample per channel.

    This function supports WAV and FLAC files compliant with CDDA
    standard (16-bit stereo LPCM, 44100 Hz). Underlying C extension
    will raise TypeError for any other audio format, or OSError when
    libsndfile can't load audio samples from the file for any reason.
    """
    return accuraterip.nframes(path)


def libsndfile_version() -> str:
    """Return libsndfile version string."""
    return accuraterip.libsndfile_version()



"""
Verify rip correctness using AccurateRip database.
"""

import argparse
import sys

from arver.disc.info import get_disc_info
from arver.rip.rip import Rip
from arver.version import version_string


def _parse_args():
    parser = argparse.ArgumentParser(
        description="""Verify a set of audio files against checksums from
        AccurateRip database. Disc TOC necessary for AccurateRip lookup is
        obtained either from a physical CD in drive (recommended if disc has
        data tracks), or from MusicBrainz disc ID (-i option). Calculation of
        AccurateRip checksums requires correct track sequence, so files must
        be specified in the correct order. Pre-gap track (HTOA) must not be
        included.""")

    parser.add_argument('rip_files',
                        nargs='+',
                        metavar='file',
                        help='audio file for calculating checksums')

    parser.add_argument('-i',
                        '--disc-id',
                        metavar='disc_id',
                        nargs='?',
                        help='get disc TOC from MusicBrainz by disc ID')

    parser.add_argument('-v', '--version', action='version', version=version_string())

    return parser.parse_args()


def main():
    args = _parse_args()
    disc = get_disc_info(args.disc_id)

    if disc is None:
        print('Failed to get disc info, exiting.')
        sys.exit(1)

    print(disc)
    print()

    disc.fetch_accuraterip_data()
    if disc.accuraterip_data is None:
        print('Failed to download AccurateRip data, exiting.')
        sys.exit(2)

    print(disc.accuraterip_data.summary())
    print()

    rip = Rip(args.rip_files)
    if len(rip) == 0:
        print('No audio files were loaded. Did you specify correct files?')
        sys.exit(3)

    verdict = rip.verify(disc)

    print()
    print(verdict.as_table())
    print()
    print(verdict.summary())


if __name__ == '__main__':
    main()



"""
Display properties of a set of audio files, including copy CRCs and
two types of AccurateRip checksums.
"""

import argparse
import sys

from arver.rip.rip import Rip
from arver.version import version_string


def _parse_args():
    parser = argparse.ArgumentParser(
        description="""Display properties and checksums of specified audio
        files. Calculation of AccurateRip checksums requires correct track
        sequence, so files must be specified in the correct order. Pre-gap
        track (HTOA) must not be included. Non-audio files and unsupported
        audio formats are quietly ignored.""")

    parser.add_argument('rip_files',
                        nargs='+',
                        metavar='file',
                        help='audio file for calculating checksums')
    parser.add_argument('-v', '--version', action='version', version=version_string())

    return parser.parse_args()


def main():
    args = _parse_args()

    rip = Rip(args.rip_files)
    if len(rip) == 0:
        print('No audio files were loaded. Did you specify correct files?')
        sys.exit(1)

    rip.calculate_checksums()
    print(rip)


if __name__ == '__main__':
    main()
